@@ -23,7 +23,6 @@
 #include <fcntl.h>
 #include <inttypes.h>
 #include <limits.h>
-#include <signal.h>
 #include <stdbool.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -32,6 +31,10 @@
 #include <time.h>
 #include <unistd.h>
 #include <wchar.h>
+#include <wctype.h>
+#if HAVE_GETTEXT
+# include <libintl.h>
+#endif
 #include "builtin.h"
 #include "exec.h"
 #include "history.h"
@@ -39,6 +42,7 @@
 #include "option.h"
 #include "path.h"
 #include "redir.h"
+#include "sig.h"
 #include "strbuf.h"
 #include "util.h"
 #include "variable.h"
@@ -820,13 +824,11 @@ void check_histfile_pid(void)
     last = &histfilepids, list = histfilepids;
     while (list != NULL) {
 	assert(list->pid > 0);
-	if (kill(list->pid, 0) < 0 && errno == ESRCH) {
-	    /* invalid process id: no such process */
+	if (!process_exists(list->pid)) {
 	    *last = list->next;
 	    free(list);
 	    list = *last;
 	} else {
-	    /* valid process id */
 	    last = &list->next, list = list->next;
 	}
     }
@@ -921,9 +923,11 @@ void finalize_history(void)
 /* Adds the specified `line' to the history.
  * Must not be called while the history is locked.
  * If `line' contains newlines, `line' is separated into multiple entries. */
+/* Only lines that contain graph-class characters are added to the history. */
 void add_history(const wchar_t *line)
 {
     maybe_init_history();
+    assert(!hist_lock);
 
     if (histfile) {
 	lock_file(fileno(histfile), F_WRLCK);
@@ -933,9 +937,11 @@ void add_history(const wchar_t *line)
 
     const wchar_t *nl = wcschr(line, L'\n');
     while ((nl = wcschr(line, L'\n')) != NULL) {
-	wchar_t *line1 = xwcsndup(line, nl - line);
-	really_add_history(line1);
-	free(line1);
+	if (nl > line) {
+	    wchar_t *line1 = xwcsndup(line, nl - line);
+	    really_add_history(line1);
+	    free(line1);
+	}
 	line = nl + 1;
     }
     really_add_history(line);
@@ -950,8 +956,19 @@ void add_history(const wchar_t *line)
  * `line' must not contain newlines.
  * `histfile' must be locked and `update_time' and `update_history' must have
  * been called. */
+/* If the `line' does not contain any graph-class characters, it is not added
+ * to the history. */
 void really_add_history(const wchar_t *line)
 {
+    /* Check if `line' contains `graph' characters */
+    for (const wchar_t *s = line; ; ) {
+	if (*s == L'\0')
+	    return;
+	if (iswgraph(*s))
+	    break;
+	s++;
+    }
+
     char *mbsline = malloc_wcstombs(line);
     if (mbsline) {
 	histentry_T *entry = new_entry(hist_next_number, now, mbsline);
@@ -1221,7 +1238,7 @@ void fc_remove_last_entry(void)
 	update_history(false);
 	remove_last_entry();
 	if (histfile) {
-	    fputws(L"c\n", histfile);
+	    fwprintf(histfile, L"c\n");
 	    fflush(histfile);
 	    lock_file(fileno(histfile), F_UNLCK);
 	}
