@@ -27,6 +27,7 @@
 #include <stdlib.h>
 #include <sys/wait.h>
 #include <unistd.h>
+#include <wctype.h>
 #if HAVE_GETTEXT
 # include <libintl.h>
 #endif
@@ -315,11 +316,13 @@ found:
     pr->pr_statuscode = status;
     if (WIFEXITED(status) || WIFSIGNALED(status))
 	pr->pr_status = JS_DONE;
-    else if (WIFSTOPPED(status))
+    if (WIFSTOPPED(status))
 	pr->pr_status = JS_STOPPED;
 #ifdef HAVE_WCONTINUED
-    else if (WIFCONTINUED(status))
+    if (WIFCONTINUED(status))
 	pr->pr_status = JS_RUNNING;
+    /* On FreeBSD, when WIFCONTINUED is true, WIFSIGNALED is also true. We must
+     * be careful about the order of these checks. */
 #endif
 
     /* decide the job status from the process status:
@@ -519,7 +522,7 @@ wchar_t *get_job_name(const job_T *job)
 /* Returns a string that describes the status of the specified process
  * such as "Running" and "Stopped(SIGTSTP)".
  * The returned string must be freed by the caller iff `*needfree' is assigned
- * true. */
+ * true, otherwise it must not be modified or freed. */
 char *get_process_status_string(const process_T *p, bool *needfree)
 {
     int status, sig;
@@ -527,7 +530,7 @@ char *get_process_status_string(const process_T *p, bool *needfree)
     switch (p->pr_status) {
     case JS_RUNNING:
 	*needfree = false;
-	return gt("Running");
+	return (char *) gt("Running");
     case JS_STOPPED:
 	*needfree = true;
 	return malloc_printf(gt("Stopped(SIG%s)"),
@@ -541,7 +544,7 @@ char *get_process_status_string(const process_T *p, bool *needfree)
 exitstatus:
 	    if (status == Exit_SUCCESS) {
 		*needfree = false;
-		return gt("Done");
+		return (char *) gt("Done");
 	    } else {
 		*needfree = true;
 		return malloc_printf(gt("Done(%d)"), status);
@@ -565,13 +568,13 @@ exitstatus:
 /* Returns a string that describes the status of the specified job
  * such as "Running" and "Stopped(SIGTSTP)".
  * The returned string must be freed by the caller iff `*needfree' is assigned
- * true. */
+ * true, otherwise it must not be modified or freed. */
 char *get_job_status_string(const job_T *job, bool *needfree)
 {
     switch (job->j_status) {
     case JS_RUNNING:
 	*needfree = false;
-	return gt("Running");
+	return (char *) gt("Running");
     case JS_STOPPED:
 	/* find a stopped process */
 	for (size_t i = job->j_pcount; ; )
