@@ -26,16 +26,20 @@
 #include <stdint.h>
 #include <stdio.h>
 #include <stdlib.h>
+#include <string.h>
+#include <sys/stat.h>
 #include <unistd.h>
 #include <wchar.h>
 #include <wctype.h>
+#if HAVE_GETTEXT
+# include <libintl.h>
+#endif
 #if YASH_ENABLE_ALIAS
-#include "../alias.h"
+# include "../alias.h"
 #endif
 #include "../exec.h"
-#if YASH_ENABLE_HISTORY
+#include "../expand.h"
 #include "../history.h"
-#endif
 #include "../job.h"
 #include "../option.h"
 #include "../path.h"
@@ -43,6 +47,7 @@
 #include "../redir.h"
 #include "../strbuf.h"
 #include "../util.h"
+#include "../wfnmatch.h"
 #include "../yash.h"
 #include "display.h"
 #include "editing.h"
@@ -56,14 +61,27 @@ xwcsbuf_T le_main_buffer;
 /* The position of the cursor on the command line. */
 /* 0 <= le_main_index <= le_main_buffer.length */
 size_t le_main_index;
-#if YASH_ENABLE_HISTORY
+
 /* The history entry that is being edited now.
  * When we're editing no history entry, `main_history_entry' is a pointer to
  * `histlist'. */
 static const histentry_T *main_history_entry;
 /* The original value of `main_history_entry', converted into a wide string. */
 static wchar_t *main_history_value;
-#endif
+
+/* The direction of currently-performed command history search. */
+enum le_search_direction le_search_direction;
+/* Supplementary buffer used in command history search. */
+/* When search is not being performed, the `contents' member is NULL. */
+xwcsbuf_T le_search_buffer;
+/* The search result for the current value of `le_search_buffer'.
+ * If there is no match, `le_search_result' is `Histlist'. */
+const histentry_T *le_search_result;
+/* The search string and the direction of the last search. */
+static struct {
+    enum le_search_direction direction;
+    wchar_t *value;
+} last_search;
 
 /* The last executed command and the currently executing command. */
 static struct command {
@@ -107,13 +125,11 @@ static plist_T undo_history;
 /* Index of the current state in the history.
  * If the current state is the newest, the index is `undo_history.length'. */
 static size_t undo_index;
-#if YASH_ENABLE_HISTORY
 /* The history entry that is saved in the undo history. */
 static const histentry_T *undo_history_entry;
 /* The index that is to be the value of the `index' member of the next undo
  * history entry. */
 static size_t undo_save_index;
-#endif
 /* Structure of history entries */
 struct undo_history {
     size_t index;        /* index of the cursor */
@@ -143,22 +159,24 @@ static void add_to_kill_ring(const wchar_t *s, size_t n)
 
 static bool alert_if_first(void);
 static bool alert_if_last(void);
-static bool alert_if_pending(void);
 static void move_cursor_forward_char(int offset);
 static void move_cursor_backward_char(int offset);
 static void move_cursor_forward_bigword(int count);
 static void move_cursor_backward_bigword(int count);
 static void move_cursor_forward_word(int count);
+static bool need_cw_treatment(void)
+    __attribute__((pure));
 static void move_cursor_backward_word(int count);
 static size_t next_bigword_index(const wchar_t *s, size_t i)
     __attribute__((nonnull));
-static size_t next_end_of_bigword_index(const wchar_t *s, size_t i)
+static size_t next_end_of_bigword_index(
+	const wchar_t *s, size_t i, bool progress)
     __attribute__((nonnull));
 static size_t previous_bigword_index(const wchar_t *s, size_t i)
     __attribute__((nonnull));
 static size_t next_word_index(const wchar_t *s, size_t i)
     __attribute__((nonnull));
-static size_t next_end_of_word_index(const wchar_t *s, size_t i)
+static size_t next_end_of_word_index(const wchar_t *s, size_t i, bool progress)
     __attribute__((nonnull));
 static size_t previous_word_index(const wchar_t *s, size_t i)
     __attribute__((nonnull));
@@ -177,17 +195,28 @@ static void vi_replace_char(wchar_t c);
 static void exec_edit_command(enum motion_expect_command cmd);
 static inline void exec_edit_command_to_eol(enum motion_expect_command cmd);
 static void vi_exec_alias(wchar_t c);
-
-#if YASH_ENABLE_HISTORY
+struct xwcsrange { const wchar_t *start, *end; };
+static struct xwcsrange get_next_bigword(const wchar_t *s)
+    __attribute__((nonnull));
+static struct xwcsrange get_prev_bigword(
+	const wchar_t *beginning, const wchar_t *s)
+    __attribute__((nonnull));
+static void search_again(enum le_search_direction dir);
 static void go_to_history_absolute(const histentry_T *e)
     __attribute__((nonnull));
 static void go_to_history_relative(int offset, bool cursorend);
 static void go_to_history(const histentry_T *e, bool cursorend)
     __attribute__((nonnull));
-#endif
+static bool update_last_search_value(void)
+    __attribute__((pure));
+static void update_search(void);
+static void perform_search(const wchar_t *pattern, enum le_search_direction dir)
+    __attribute__((nonnull));
 
-#define ALERT_AND_RETURN_IF_PENDING \
-    do if (alert_if_pending()) return; while (0)
+#define ALERT_AND_RETURN_IF_PENDING                     \
+    do if (state.pending_command_motion != MEC_NONE)    \
+	{ cmd_alert(L'\0'); return; }                   \
+    while (0)
 
 
 /* Initializes the editing module before starting editing. */
@@ -195,10 +224,8 @@ void le_editing_init(void)
 {
     wb_init(&le_main_buffer);
     le_main_index = 0;
-#if YASH_ENABLE_HISTORY
     main_history_entry = Histlist;
     main_history_value = xwcsdup(L"");
-#endif
 
     switch (shopt_lineedit) {
 	case shopt_vi:
@@ -212,16 +239,11 @@ void le_editing_init(void)
     last_command.func = 0;
     last_command.arg = L'\0';
 
-#if YASH_ENABLE_HISTORY
     start_using_history();
-#endif
-
     pl_init(&undo_history);
     undo_index = 0;
-#if YASH_ENABLE_HISTORY
     undo_save_index = le_main_index;
     undo_history_entry = Histlist;
-#endif
     save_undo_history();
 
     reset_state();
@@ -232,12 +254,12 @@ void le_editing_init(void)
  * Returns the content of the main buffer, which must be freed by the caller. */
 wchar_t *le_editing_finalize(void)
 {
+    assert(le_search_buffer.contents == NULL);
+
     recfree(pl_toary(&undo_history), free);
 
-#if YASH_ENABLE_HISTORY
     end_using_history();
     free(main_history_value);
-#endif
 
     wb_wccat(&le_main_buffer, L'\n');
     return wb_towcs(&le_main_buffer);
@@ -329,9 +351,7 @@ void save_undo_history(void)
     wcscpy(e->contents, le_main_buffer.contents);
     pl_add(&undo_history, e);
     assert(undo_index == undo_history.length - 1);
-#if YASH_ENABLE_HISTORY
     undo_history_entry = main_history_entry;
-#endif
 }
 
 /* Calls `save_undo_history' if the current contents of the edit line is not
@@ -339,10 +359,9 @@ void save_undo_history(void)
 void maybe_save_undo_history(void)
 {
     assert(undo_index <= undo_history.length);
-#if YASH_ENABLE_HISTORY
+    size_t save_undo_save_index = undo_save_index;
     undo_save_index = le_main_index;
     if (undo_history_entry == main_history_entry) {
-#endif
 	if (undo_index < undo_history.length) {
 	    struct undo_history *h = undo_history.contents[undo_index];
 	    if (wcscmp(le_main_buffer.contents, h->contents) == 0) {
@@ -351,7 +370,6 @@ void maybe_save_undo_history(void)
 	    }
 	    undo_index++;
 	}
-#if YASH_ENABLE_HISTORY
     } else {
 	if (wcscmp(le_main_buffer.contents, main_history_value) == 0)
 	    return;
@@ -364,12 +382,12 @@ void maybe_save_undo_history(void)
 	pl_clear(&undo_history, free);
 	h = xmalloc(sizeof *h +
 		(wcslen(main_history_value) + 1) * sizeof *h->contents);
-	h->index = undo_save_index;
+	assert(save_undo_save_index <= wcslen(main_history_value));
+	h->index = save_undo_save_index;
 	wcscpy(h->contents, main_history_value);
 	pl_add(&undo_history, h);
 	undo_index = 1;
     }
-#endif /* YASH_ENABLE_HISTORY */
     save_undo_history();
 }
 
@@ -457,7 +475,7 @@ bool alert_if_first(void)
     if (le_main_index > 0)
 	return false;
 
-    cmd_alert(L'\a');
+    cmd_alert(L'\0');
     return true;
 }
 
@@ -466,6 +484,8 @@ bool alert_if_first(void)
 bool alert_if_last(void)
 {
     if (le_current_mode == &le_modes[LE_MODE_VI_COMMAND]) {
+	if (state.pending_command_motion != MEC_NONE)
+	    return false;
 	if (le_main_buffer.length > 0
 		&& le_main_index < le_main_buffer.length - 1)
 	    return false;
@@ -474,21 +494,10 @@ bool alert_if_last(void)
 	    return false;
     }
 
-    cmd_alert(L'\a');
+    cmd_alert(L'\0');
     return true;
 }
 
-/* Invokes `cmd_alert' and returns true if the pending command is set. */
-bool alert_if_pending(void)
-{
-    if (state.pending_command_motion != MEC_NONE) {
-	cmd_alert(L'\a');
-	return true;
-    } else {
-	return false;
-    }
-}
-
 /* Inserts one character in the buffer.
  * If the count is set, inserts `count' times.
  * If `overwrite' is true, overwrites the character instead of inserting. */
@@ -509,7 +518,7 @@ void cmd_self_insert(wchar_t c)
 		!overwrite && le_main_index == le_main_buffer.length);
 	reset_state();
     } else {
-	cmd_alert(c);
+	cmd_alert(L'\0');
     }
 }
 
@@ -539,7 +548,7 @@ void cmd_digit_argument(wchar_t c)
 {
     if (L'0' <= c && c <= L'9') {
 	if (state.count.abs > COUNT_ABS_MAX / 10) {
-	    cmd_alert(c);  // argument too large
+	    cmd_alert(L'\0');  // argument too large
 	    return;
 	}
 	if (state.count.sign == 0)
@@ -640,7 +649,8 @@ void cmd_end_of_bigword(wchar_t c __attribute__((unused)))
     int count = get_count(1);
     size_t new_index = le_main_index;
     while (--count >= 0 && new_index < le_main_buffer.length)
-	new_index = next_end_of_bigword_index(le_main_buffer.contents,new_index);
+	new_index = next_end_of_bigword_index(
+		le_main_buffer.contents, new_index, true);
     exec_motion_command(new_index, true);
 }
 
@@ -663,9 +673,21 @@ void move_cursor_forward_bigword(int count)
 	return;
 
     size_t new_index = le_main_index;
-    while (--count >= 0 && new_index < le_main_buffer.length)
-	new_index = next_bigword_index(le_main_buffer.contents, new_index);
-    exec_motion_command(new_index, false);
+    if (!need_cw_treatment()) {
+	while (--count >= 0 && new_index < le_main_buffer.length)
+	    new_index = next_bigword_index(le_main_buffer.contents, new_index);
+	exec_motion_command(new_index, false);
+    } else {
+	while (count > 1 && new_index < le_main_buffer.length) {
+	    new_index = next_bigword_index(le_main_buffer.contents, new_index);
+	    count--;
+	}
+	if (count > 0 && new_index < le_main_buffer.length) {
+	    new_index = next_end_of_bigword_index(
+		    le_main_buffer.contents, new_index, false);
+	}
+	exec_motion_command(new_index, true);
+    }
 }
 
 /* Moves the cursor backward `count' bigwords, relative to the current position.
@@ -698,7 +720,7 @@ size_t next_bigword_index(const wchar_t *s, size_t i)
  * index `i'. If `i' is at the end of a bigword, the end of the next bigword is
  * returned. The return value is greater than `i' unless `s[i]' is a null
  * character. */
-size_t next_end_of_bigword_index(const wchar_t *s, size_t i)
+size_t next_end_of_bigword_index(const wchar_t *s, size_t i, bool progress)
 {
     const size_t init = i;
 start:
@@ -709,7 +731,7 @@ start:
     while (s[i] != L'\0' && !iswblank(s[i]))
 	i++;
     i--;
-    if (i > init) {
+    if (i > init || !progress) {
 	return i;
     } else {
 	i++;
@@ -760,7 +782,8 @@ void cmd_end_of_word(wchar_t c __attribute__((unused)))
     int count = get_count(1);
     size_t new_index = le_main_index;
     while (--count >= 0 && new_index < le_main_buffer.length)
-	new_index = next_end_of_word_index(le_main_buffer.contents, new_index);
+	new_index = next_end_of_word_index(
+		le_main_buffer.contents, new_index, true);
     exec_motion_command(new_index, true);
 }
 
@@ -783,9 +806,33 @@ void move_cursor_forward_word(int count)
 	return;
 
     size_t new_index = le_main_index;
-    while (--count >= 0 && new_index < le_main_buffer.length)
-	new_index = next_word_index(le_main_buffer.contents, new_index);
-    exec_motion_command(new_index, false);
+    if (!need_cw_treatment()) {
+	while (--count >= 0 && new_index < le_main_buffer.length)
+	    new_index = next_word_index(le_main_buffer.contents, new_index);
+	exec_motion_command(new_index, false);
+    } else {
+	while (count > 1 && new_index < le_main_buffer.length) {
+	    new_index = next_word_index(le_main_buffer.contents, new_index);
+	    count--;
+	}
+	if (count > 0 && new_index < le_main_buffer.length) {
+	    new_index = next_end_of_word_index(
+		    le_main_buffer.contents, new_index, false);
+	}
+	exec_motion_command(new_index, true);
+    }
+}
+
+/* Checks if we need a special treatment for the "cw" and "cW" commands. */
+bool need_cw_treatment(void)
+{
+    switch (state.pending_command_motion) {
+	case MEC_CHANGE:
+	case MEC_COPYCHANGE:
+	    return !iswblank(le_main_buffer.contents[le_main_index]);
+	default:
+	    return false;
+    }
 }
 
 /* Moves the cursor backward `count' words, relative to the current position.
@@ -831,9 +878,13 @@ size_t next_word_index(const wchar_t *s, size_t i)
 }
 
 /* Returns the index of the end of a word in the string `s', counted from index
- * `i'. If `i' is at the end of a word, the end of the next word is returned.
- * The return value is greater than `i' unless `s[i]' is a null character. */
-size_t next_end_of_word_index(const wchar_t *s, size_t i)
+ * `i'.
+ * If `progress' is true:
+ *   If `i' is at the end of a word, the end of the next word is returned.
+ *   The return value is greater than `i' unless `s[i]' is a null character.
+ * If `progress' is false:
+ *   If `i' is at the end of a word, `i' is returned. */
+size_t next_end_of_word_index(const wchar_t *s, size_t i, bool progress)
 {
     const size_t init = i;
 start:
@@ -848,10 +899,11 @@ start:
     } else {
 	do
 	    i++;
-	while (!iswblank(s[i]) && !(s[i] == L'_' || iswalnum(s[i])));
+	while (s[i] != L'\0' && s[i] != L'_'
+		&& !iswblank(s[i]) && !iswalnum(s[i]));
     }
     i--;
-    if (i > init) {
+    if (i > init || !progress) {
 	return i;
     } else {
 	i++;
@@ -878,7 +930,7 @@ start:
 	    if (i == 0)
 		return 0;
 	    i--;
-	} while (!iswblank(s[i]) && !(s[i] == L'_' || iswalnum(s[i])));
+	} while (s[i] != L'_' && !iswblank(s[i]) && !iswalnum(s[i]));
     }
     i++;
     if (i < init) {
@@ -930,6 +982,7 @@ void cmd_accept_line(wchar_t c __attribute__((unused)))
 {
     ALERT_AND_RETURN_IF_PENDING;
 
+    cmd_srch_accept_search(L'\0');
     le_editstate = LE_EDITSTATE_DONE;
     reset_state();
 }
@@ -938,13 +991,14 @@ void cmd_accept_line(wchar_t c __attribute__((unused)))
  * `le_state' is set to LE_STATE_INTERRUPTED and `le_readline' returns. */
 void cmd_abort_line(wchar_t c __attribute__((unused)))
 {
+    cmd_srch_abort_search(L'\0');
     le_editstate = LE_EDITSTATE_INTERRUPTED;
     reset_state();
 }
 
 /* If the edit line is empty, sets `le_state' to LE_STATE_ERROR (return EOF).
  * Otherwise, causes alert. */
-void cmd_eof_if_empty(wchar_t c)
+void cmd_eof_if_empty(wchar_t c __attribute__((unused)))
 {
     ALERT_AND_RETURN_IF_PENDING;
 
@@ -952,7 +1006,7 @@ void cmd_eof_if_empty(wchar_t c)
 	le_editstate = LE_EDITSTATE_ERROR;
 	reset_state();
     } else {
-	cmd_alert(c);
+	cmd_alert(L'\0');
     }
 }
 
@@ -1011,10 +1065,18 @@ void cmd_expect_char(wchar_t c)
 {
     if (state.pending_command_char) {
 	current_command.func = state.pending_command_char;
+	current_command.arg = c;
 	state.pending_command_char(c);
     }
 }
 
+/* Executes a command that expects a character as an argument with L'\\'
+ * given as the argument. */
+void cmd_expect_backslash(wchar_t c __attribute__((unused)))
+{
+    cmd_expect_char(L'\\');
+}
+
 /* Cancels a command that expects a character as an argument. */
 void cmd_abort_expect_char(wchar_t c __attribute__((unused)))
 {
@@ -1026,7 +1088,10 @@ void cmd_abort_expect_char(wchar_t c __attribute__((unused)))
 void cmd_redraw_all(wchar_t c __attribute__((unused)))
 {
     le_display_clear();
-    le_display_print_all();
+    le_restore_terminal();
+    le_setupterm();
+    le_set_terminal();
+    le_display_print_all(false);
 }
 
 
@@ -1154,7 +1219,7 @@ void cmd_backward_delete_line(wchar_t c __attribute__((unused)))
 
 /* Kills the character under the cursor.
  * If the count is set, `count' characters are killed. */
-void cmd_kill_char(wchar_t c)
+void cmd_kill_char(wchar_t c __attribute__((unused)))
 {
     ALERT_AND_RETURN_IF_PENDING;
     save_current_edit_command();
@@ -1162,7 +1227,7 @@ void cmd_kill_char(wchar_t c)
 
     assert(le_main_index <= le_main_buffer.length);
     if (le_main_index == le_main_buffer.length) {
-	cmd_alert(c);
+	cmd_alert(L'\0');
 	return;
     }
 
@@ -1172,7 +1237,7 @@ void cmd_kill_char(wchar_t c)
 /* Kills the character behind the cursor.
  * If the count is set, `count' characters are killed.
  * If the cursor is at the beginning of the line, the terminal is alerted. */
-void cmd_backward_kill_char(wchar_t c)
+void cmd_backward_kill_char(wchar_t c __attribute__((unused)))
 {
     ALERT_AND_RETURN_IF_PENDING;
     save_current_edit_command();
@@ -1180,7 +1245,7 @@ void cmd_backward_kill_char(wchar_t c)
 
     assert(le_main_index <= le_main_buffer.length);
     if (le_main_index == 0) {
-	cmd_alert(c);
+	cmd_alert(L'\0');
 	return;
     }
 
@@ -1259,7 +1324,7 @@ void insert_killed_string(bool after_cursor, bool cursor_on_last_char)
 
     const wchar_t *s = kill_ring[last_put_elem];
     if (s == NULL) {
-	cmd_alert(L'\a');
+	cmd_alert(L'\0');
 	return;
     } else if (s[0] == L'\0') {
 	reset_state();
@@ -1286,22 +1351,22 @@ void insert_killed_string(bool after_cursor, bool cursor_on_last_char)
 
 /* Replaces the string just inserted by `cmd_put_left' with the previously
  * killed string. */
-void cmd_put_pop(wchar_t c)
+void cmd_put_pop(wchar_t c __attribute__((unused)))
 {
     ALERT_AND_RETURN_IF_PENDING;
     if (last_command.func != cmd_put_left
 	    && last_command.func != cmd_put
 	    && last_command.func != cmd_put_before) {
-	cmd_alert(c);
+	cmd_alert(L'\0');
 	return;
     }
     const wchar_t *s = kill_ring[last_put_elem];
     if (s == NULL || s[0] == L'\0') {
-	cmd_alert(c);
+	cmd_alert(L'\0');
 	return;
     }
     save_current_edit_command();
-    //TODO
+    //TODO: cmd_put_pop
 }
 
 /* Undoes the last editing command. */
@@ -1370,17 +1435,17 @@ void cancel_undo(int offset)
     return;
 
 error:
-    cmd_alert(L'\a');
+    cmd_alert(L'\0');
     return;
 }
 
 /* Redoes the last editing command. */
 /* XXX: currently vi's "i" command cannot be redone. */
-void cmd_redo(wchar_t c)
+void cmd_redo(wchar_t c __attribute__((unused)))
 {
     ALERT_AND_RETURN_IF_PENDING;
     if (!last_edit_command.command.func) {
-	cmd_alert(c);
+	cmd_alert(L'\0');
 	return;
     }
 
@@ -1505,7 +1570,7 @@ void exec_find(wchar_t c, int count, bool till)
     return;
 
 error:
-    cmd_alert(c);
+    cmd_alert(L'\0');
     return;
 }
 
@@ -1539,10 +1604,10 @@ size_t find_nth_occurence(wchar_t c, int n)
 }
 
 /* Redoes the last find/till command. */
-void cmd_vi_refind(wchar_t c)
+void cmd_vi_refind(wchar_t c __attribute__((unused)))
 {
     if (!last_find_command.func) {
-	cmd_alert(c);
+	cmd_alert(L'\0');
 	return;
     }
 
@@ -1550,10 +1615,10 @@ void cmd_vi_refind(wchar_t c)
 }
 
 /* Redoes the last find/till command in the reverse direction. */
-void cmd_vi_refind_rev(wchar_t c)
+void cmd_vi_refind_rev(wchar_t c __attribute__((unused)))
 {
     if (!last_find_command.func) {
-	cmd_alert(c);
+	cmd_alert(L'\0');
 	return;
     }
 
@@ -1596,7 +1661,7 @@ void vi_replace_char(wchar_t c)
 	le_display_reprint_buffer(old_index, false);
 	reset_state();
     } else {
-	cmd_alert(c);
+	cmd_alert(L'\0');
     }
 }
 
@@ -1651,11 +1716,11 @@ void cmd_vi_change_case(wchar_t c)
     size_t old_index = le_main_index;
 
     if (le_main_index == le_main_buffer.length) {
-	cmd_alert(c);
+	cmd_alert(L'\0');
 	return;
     }
     for (int count = get_count(1); --count >= 0; ) {
-	wchar_t c = le_main_buffer.contents[le_main_index];
+	c = le_main_buffer.contents[le_main_index];
 	le_main_buffer.contents[le_main_index]
 	    = (iswlower(c) ? towupper : towlower)(c);
 	le_main_index++;
@@ -1760,7 +1825,7 @@ void exec_edit_command(enum motion_expect_command cmd)
 	    if (old_index <= le_main_buffer.length)
 		le_main_index = old_index;
 	} else {
-	    cmd_alert(L'\a');
+	    cmd_alert(L'\0');
 	}
     } else {
 	state.count.multiplier = get_count(1);
@@ -1782,7 +1847,7 @@ void exec_edit_command_to_eol(enum motion_expect_command cmd)
 
 /* Kills the character under the cursor and sets the editing mode to
  * "vi insert". If the count is set, `count' characters are killed. */
-void cmd_vi_substitute(wchar_t c)
+void cmd_vi_substitute(wchar_t c __attribute__((unused)))
 {
     ALERT_AND_RETURN_IF_PENDING;
     save_current_edit_command();
@@ -1790,7 +1855,7 @@ void cmd_vi_substitute(wchar_t c)
 
     assert(le_main_index <= le_main_buffer.length);
     if (le_main_index == 0) {
-	cmd_alert(c);
+	cmd_alert(L'\0');
 	return;
     }
 
@@ -1799,6 +1864,101 @@ void cmd_vi_substitute(wchar_t c)
     overwrite = false;
 }
 
+/* Appends a space followed by the last bigword from the newest history entry.
+ * If the count is specified, the `count'th word is appended.
+ * The mode is changed to vi-insert. */
+void cmd_vi_append_last_bigword(wchar_t c __attribute__((unused)))
+{
+    ALERT_AND_RETURN_IF_PENDING;
+    save_current_edit_command();
+    maybe_save_undo_history();
+
+    wchar_t *lastcmd = NULL;
+    int count = get_count(-1);
+    if (count == 0 || histlist.count == 0)
+	goto fail;
+
+    struct xwcsrange range;
+    lastcmd = malloc_mbstowcs(histlist.Newest->value);
+    if (lastcmd == NULL)
+	goto fail;
+    if (count >= 0) {
+	/* find the count'th word */
+	range.start = range.end = lastcmd;
+	do {
+	    struct xwcsrange r = get_next_bigword(range.end);
+	    if (r.start == r.end)
+		break;
+	    range = r;
+	} while (--count > 0 && *range.end != L'\0');
+    } else {
+	/* find the count'th last word */
+	range.start = range.end = lastcmd + wcslen(lastcmd);
+	do {
+	    struct xwcsrange r = get_prev_bigword(lastcmd, range.start);
+	    if (r.start == r.end)
+		break;
+	    range = r;
+	} while (++count < 0 && lastcmd < range.start);
+    }
+    assert(range.start <= range.end);
+    if (range.start == range.end)
+	goto fail;
+
+    if (le_main_index < le_main_buffer.length)
+	le_main_index++;
+    size_t oldindex = le_main_index;
+    size_t len = range.end - range.start;
+    wb_ninsert_force(&le_main_buffer, le_main_index, L" ", 1);
+    le_main_index += 1;
+    wb_ninsert_force(&le_main_buffer, le_main_index, range.start, len);
+    le_main_index += len;
+    free(lastcmd);
+    le_display_reprint_buffer(oldindex, le_main_index == le_main_buffer.length);
+    cmd_setmode_viinsert(L'\0');
+    return;
+
+fail:
+    free(lastcmd);
+    cmd_alert(L'\0');
+    return;
+}
+
+struct xwcsrange get_next_bigword(const wchar_t *s)
+{
+    struct xwcsrange result;
+    while (iswblank(*s))
+	s++;
+    result.start = s;
+    while (*s != L'\0' && !iswblank(*s))
+	s++;
+    result.end = s;
+    return result;
+    /* result.start == result.end if no bigword found */
+}
+
+struct xwcsrange get_prev_bigword(const wchar_t *beginning, const wchar_t *s)
+{
+    struct xwcsrange result;
+    assert(beginning <= s);
+    do {
+	if (beginning == s) {
+	    result.start = result.end = s;
+	    return result;
+	}
+    } while (iswblank(*--s));
+    result.end = s + 1;
+    do {
+	if (beginning == s) {
+	    result.start = s;
+	    return result;
+	}
+    } while (!iswblank(*--s));
+    result.start = s + 1;
+    return result;
+    /* result.start == result.end if no bigword found */
+}
+
 /* Sets the editing mode to "vi expect" and the pending command to
  * `vi_exec_alias'. */
 void cmd_vi_exec_alias(wchar_t c __attribute__((unused)))
@@ -1828,13 +1988,13 @@ void vi_exec_alias(wchar_t c)
 	}
     }
 #endif
-    cmd_alert(c);
+    cmd_alert(L'\0');
 }
 
 /* Invokes an external command to edit the current line and accepts the result.
  * If the editor returns a non-zero status, the line is not accepted. */
 /* cf. history.c:fc_edit_and_exec_entries */
-void cmd_vi_edit_and_accept(wchar_t c)
+void cmd_vi_edit_and_accept(wchar_t c __attribute__((unused)))
 {
     ALERT_AND_RETURN_IF_PENDING;
 
@@ -1846,13 +2006,13 @@ void cmd_vi_edit_and_accept(wchar_t c)
 
     fd = create_temporary_file(&tempfile, S_IRUSR | S_IWUSR);
     if (fd < 0) {
-	cmd_alert(c);
+	cmd_alert(L'\0');
 	return;
     }
     f = fdopen(fd, "w");
     if (f == NULL) {
 	xclose(fd);
-	cmd_alert(c);
+	cmd_alert(L'\0');
 	return;
     }
     le_suspend_readline();
@@ -1863,8 +2023,8 @@ void cmd_vi_edit_and_accept(wchar_t c)
 	fclose(f);
 	unlink(tempfile);
 	free(tempfile);
-	le_display_print_all();
-	cmd_alert(c);
+	le_resume_readline();
+	cmd_alert(L'\0');
     } else if (cpid > 0) {  // parent process
 	fclose(f);
 
@@ -1878,7 +2038,7 @@ void cmd_vi_edit_and_accept(wchar_t c)
 
 	f = fopen(tempfile, "r");
 	if (f == NULL) {
-	    cmd_alert(c);
+	    cmd_alert(L'\0');
 	} else {
 	    wint_t c;
 
@@ -1904,8 +2064,7 @@ end:
 	free(tempfile);
 	le_resume_readline();
     } else {  // child process
-	fputws(le_main_buffer.contents, f);
-	fputwc(L'\n', f);
+	fwprintf(f, L"%ls\n");
 	fclose(f);
 
 	wchar_t *command = malloc_wprintf(L"vi %s", tempfile);
@@ -1918,6 +2077,72 @@ end:
     }
 }
 
+/* Starts vi-like command history search in the forward direction. */
+void cmd_vi_search_forward(wchar_t c __attribute__((unused)))
+{
+    ALERT_AND_RETURN_IF_PENDING;
+
+    le_search_direction = FORWARD;
+    wb_init(&le_search_buffer);
+    le_set_mode(LE_MODE_VI_SEARCH);
+    update_search();
+}
+
+/* Starts vi-like command history search in the backward direction. */
+void cmd_vi_search_backward(wchar_t c __attribute__((unused)))
+{
+    ALERT_AND_RETURN_IF_PENDING;
+
+    le_search_direction = BACKWARD;
+    wb_init(&le_search_buffer);
+    le_set_mode(LE_MODE_VI_SEARCH);
+    update_search();
+}
+
+/* Redoes the last vi-like search. */
+void cmd_vi_search_again(wchar_t c __attribute__((unused)))
+{
+    search_again(last_search.direction);
+}
+
+/* Redoes the last vi-like search in the reverse direction. */
+void cmd_vi_search_again_rev(wchar_t c __attribute__((unused)))
+{
+    switch (last_search.direction) {
+	case FORWARD:   search_again(BACKWARD);  break;
+	case BACKWARD:  search_again(FORWARD);   break;
+    }
+}
+
+/* Redoes the last vi-like search in the forward direction. */
+void cmd_vi_search_again_forward(wchar_t c __attribute__((unused)))
+{
+    search_again(FORWARD);
+}
+
+/* Redoes the last vi-like search in the backward direction. */
+void cmd_vi_search_again_backward(wchar_t c __attribute__((unused)))
+{
+    search_again(BACKWARD);
+}
+
+void search_again(enum le_search_direction dir)
+{
+    ALERT_AND_RETURN_IF_PENDING;
+
+    if (last_search.value == NULL) {
+	cmd_alert(L'\0');
+	return;
+    }
+
+    perform_search(last_search.value, dir);
+    if (le_search_result == Histlist) {
+	cmd_alert(L'\0');
+    } else {
+	go_to_history(le_search_result, false);
+    }
+}
+
 
 /********** History-Related Commands **********/
 
@@ -1925,46 +2150,29 @@ end:
  * If the `count' is specified, goes to the history entry whose number is
  * `count'. If the specified entry is not found, make an alert.
  * The cursor is put at the beginning of line. */
-void cmd_oldest_history(wchar_t c)
+void cmd_oldest_history(wchar_t c __attribute__((unused)))
 {
-#if YASH_ENABLE_HISTORY
-    (void) c;
     go_to_history_absolute(histlist.Oldest);
-#else
-    cmd_alert(c);
-#endif
 }
 
 /* Goes to the newest history entry.
  * If the `count' is specified, goes to the history entry whose number is
  * `count'. If the specified entry is not found, make an alert.
  * The cursor is put at the beginning of line. */
-void cmd_newest_history(wchar_t c)
+void cmd_newest_history(wchar_t c __attribute__((unused)))
 {
-#if YASH_ENABLE_HISTORY
-    (void) c;
     go_to_history_absolute(histlist.Newest);
-#else
-    cmd_alert(c);
-#endif
 }
 
 /* Goes to the newest history entry.
  * If the `count' is specified, goes to the history entry whose number is
  * `count'. If the specified entry is not found, make an alert.
  * The cursor is put at the beginning of line. */
-void cmd_return_history(wchar_t c)
+void cmd_return_history(wchar_t c __attribute__((unused)))
 {
-#if YASH_ENABLE_HISTORY
-    (void) c;
     go_to_history_absolute(Histlist);
-#else
-    cmd_alert(c);
-#endif
 }
 
-#if YASH_ENABLE_HISTORY
-
 /* Goes to the specified history entry `e'.
  * If the `count' is specified, goes to the history entry whose number is
  * `count'. If the specified entry is not found, make an alert.
@@ -1988,66 +2196,42 @@ void go_to_history_absolute(const histentry_T *e)
     return;
 
 alert:
-    cmd_alert(L'\a');
+    cmd_alert(L'\0');
     return;
 }
 
-#endif /* YASH_ENABLE_HISTORY */
-
 /* Goes to the `count'th next history entry.
  * The cursor is put at the beginning of line. */
-void cmd_next_history(wchar_t c)
+void cmd_next_history(wchar_t c __attribute__((unused)))
 {
-#if YASH_ENABLE_HISTORY
-    (void) c;
     ALERT_AND_RETURN_IF_PENDING;
     go_to_history_relative(get_count(1), false);
-#else
-    cmd_alert(c);
-#endif
 }
 
 /* Goes to the `count'th previous history entry.
  * The cursor is put at the beginning of line. */
-void cmd_prev_history(wchar_t c)
+void cmd_prev_history(wchar_t c __attribute__((unused)))
 {
-#if YASH_ENABLE_HISTORY
-    (void) c;
     ALERT_AND_RETURN_IF_PENDING;
     go_to_history_relative(-get_count(1), false);
-#else
-    cmd_alert(c);
-#endif
 }
 
 /* Goes to the `count'th next history entry.
  * The cursor is put at the end of line. */
-void cmd_next_history_eol(wchar_t c)
+void cmd_next_history_eol(wchar_t c __attribute__((unused)))
 {
-#if YASH_ENABLE_HISTORY
-    (void) c;
     ALERT_AND_RETURN_IF_PENDING;
     go_to_history_relative(get_count(1), true);
-#else
-    cmd_alert(c);
-#endif
 }
 
 /* Goes to the `count'th previous history entry.
  * The cursor is put at the end of line. */
-void cmd_prev_history_eol(wchar_t c)
+void cmd_prev_history_eol(wchar_t c __attribute__((unused)))
 {
-#if YASH_ENABLE_HISTORY
-    (void) c;
     ALERT_AND_RETURN_IF_PENDING;
     go_to_history_relative(-get_count(1), true);
-#else
-    cmd_alert(c);
-#endif
 }
 
-#if YASH_ENABLE_HISTORY
-
 /* Goes to the `offset'th next history entry.
  * If `cursorend' is true, the cursor is put at the end of line; otherwise, at
  * the beginning of line. */
@@ -2070,7 +2254,7 @@ void go_to_history_relative(int offset, bool cursorend)
     go_to_history(e, cursorend);
     return;
 alert:
-    cmd_alert(L'\a');
+    cmd_alert(L'\0');
     return;
 }
 
@@ -2102,7 +2286,188 @@ void go_to_history(const histentry_T *e, bool cursorend)
     reset_state();
 }
 
-#endif /* YASH_ENABLE_HISTORY */
+/***** History Search Commands *****/
+
+/* Appends the argument character to the search buffer. */
+void cmd_srch_self_insert(wchar_t c)
+{
+    if (le_search_buffer.contents == NULL || c == L'\0') {
+	cmd_alert(L'\0');
+	return;
+    }
+
+    wb_wccat(&le_search_buffer, c);
+    update_search();
+}
+
+/* Appends a backslash to the search buffer. */
+void cmd_srch_insert_backslash(wchar_t c __attribute__((unused)))
+{
+    cmd_srch_self_insert(L'\\');
+}
+
+/* Removes the last character from the search buffer.
+ * If there are no characters in the buffer, calls `cmd_srch_abort_search' (for
+ * vi-like search) or `cmd_alert' (for emacs-like search). */
+void cmd_srch_backward_delete_char(wchar_t c __attribute__((unused)))
+{
+    if (le_search_buffer.contents == NULL) {
+	cmd_alert(L'\0');
+	return;
+    }
+
+    // XXX currently, no emacs-like search
+    if (le_search_buffer.length == 0) {
+	cmd_srch_abort_search(L'\0');
+	return;
+    }
+
+    wb_remove(&le_search_buffer, le_search_buffer.length - 1, 1);
+    update_search();
+}
+
+/* Finishes the history search with the current result candicate.
+ * If no search is being performed, does nothing. */
+void cmd_srch_accept_search(wchar_t c __attribute__((unused)))
+{
+    if (le_search_buffer.contents == NULL)
+	return;
+
+    last_search.direction = le_search_direction;
+    if (update_last_search_value()) {
+	free(last_search.value);
+	last_search.value = wb_towcs(&le_search_buffer);
+    } else {
+	wb_destroy(&le_search_buffer);
+    }
+    le_search_buffer.contents = NULL;
+    le_set_mode(LE_MODE_VI_COMMAND); // XXX assumes vi
+    if (le_search_result == Histlist) {
+	cmd_alert(L'\0');
+	le_display_reprint_buffer(0, false);
+    } else {
+	go_to_history(le_search_result, false);
+	// XXX cursor should be at end for emacs-like search
+    }
+}
+
+/* Checks if we should update `last_search.value' to the current value of
+ * `le_search_buffer'. */
+bool update_last_search_value(void)
+{
+    // XXX assumes vi
+    if (le_search_buffer.contents[0] == L'\0')
+	return false;
+    if (le_search_buffer.contents[0] == L'^'
+	    && le_search_buffer.contents[1] == L'\0')
+	return false;
+    return true;
+}
+
+/* Aborts the history search.
+ * If no search is being performed, does nothing. */
+void cmd_srch_abort_search(wchar_t c __attribute__((unused)))
+{
+    if (le_search_buffer.contents == NULL)
+	return;
+
+    wb_destroy(&le_search_buffer);
+    le_search_buffer.contents = NULL;
+    le_set_mode(LE_MODE_VI_COMMAND); // XXX assumes vi
+    le_display_reprint_buffer(0, false);
+    reset_state();
+}
+
+/* Re-calculates the search result candidate and prints it. */
+void update_search(void)
+{
+    // XXX currently no emacs-like search
+
+    const wchar_t *pattern = le_search_buffer.contents;
+    if (pattern[0] == L'\0') {
+	pattern = last_search.value;
+	if (pattern == NULL) {
+	    le_search_result = Histlist;
+	    goto done;
+	}
+    }
+
+    perform_search(pattern, le_search_direction);
+done:
+    le_display_reprint_buffer(0, false);
+    reset_state();
+}
+
+/* Performs history search with the given parameters and updates the result
+ * candidate. */
+void perform_search(const wchar_t *pattern, enum le_search_direction dir)
+{
+    const histentry_T *e = main_history_entry;
+    char *lpattern = NULL;
+    size_t minlen = minlen;
+    bool beginning;
+
+    if (dir == FORWARD && e == Histlist)
+	goto done;
+
+    if (pattern[0] == L'^') {
+	beginning = true;
+	pattern++;
+	if (pattern[0] == L'\0') {
+	    e = Histlist;
+	    goto done;
+	}
+    } else {
+	beginning = false;
+    }
+
+    if (pattern_has_special_char(pattern, false)) {
+	minlen = shortest_match_length(pattern, 0);
+    } else {
+	lpattern = realloc_wcstombs(unescape(pattern));
+	if (lpattern == NULL)
+	    goto done;
+    }
+
+    for (;;) {
+	switch (dir) {
+	    case FORWARD:   e = e->Next;  break;
+	    case BACKWARD:  e = e->Prev;  break;
+	}
+	if (e == Histlist)
+	    goto done;
+
+	if (lpattern != NULL) {
+	    if ((beginning ? matchstrprefix : strstr)(e->value, lpattern))
+		goto done;
+	} else {
+	    wchar_t *wvalue = malloc_mbstowcs(e->value);
+	    if (wvalue != NULL) {
+		size_t r;
+		if (beginning) {
+		    r = wfnmatchl(pattern, wvalue, 0, WFNM_SHORTEST, minlen);
+		} else {
+		    const wchar_t *w = wvalue;
+		    assert(w[0] != L'\0');
+		    do {
+			r = wfnmatchl(pattern, w, 0, WFNM_SHORTEST, minlen);
+			if (r != WFNM_NOMATCH)
+			    break;
+		    } while (*++w != L'\0');
+		}
+		free(wvalue);
+		switch (r) {
+		    case WFNM_NOMATCH:  break;
+		    case WFNM_ERROR:    e = Histlist;  goto done;
+		    default:            goto done;
+		}
+	    }
+	}
+    }
+done:
+    le_search_result = e;
+    free(lpattern);
+}
 
 
 /* vim: set ts=8 sts=4 sw=4 noet: */
