@@ -22,10 +22,21 @@
 #include <stddef.h>
 
 
+enum shopt_lineedit_T {
+    shopt_nolineedit,
+    shopt_vi,
+    shopt_emacs,
+};
+enum shopt_yesnoauto_T {
+    shopt_yes,
+    shopt_no,
+    shopt_auto,
+};
+
 extern _Bool posixly_correct;
 extern _Bool is_login_shell;
 extern _Bool is_interactive, is_interactive_now;
-extern _Bool do_job_control, shopt_notify;
+extern _Bool do_job_control, shopt_notify, shopt_notifyle;
 extern _Bool shopt_read_arg, shopt_read_stdin;
 extern const wchar_t *command_name;
 extern _Bool shopt_allexport, shopt_hashondef;
@@ -35,13 +46,18 @@ extern _Bool shopt_noglob, shopt_nocaseglob, shopt_dotglob, shopt_markdirs,
        shopt_extendedglob, shopt_nullglob;
 extern _Bool shopt_braceexpand;
 extern _Bool shopt_noclobber;
+extern enum shopt_lineedit_T shopt_lineedit;
+extern enum shopt_yesnoauto_T shopt_le_convmeta;
+extern _Bool shopt_le_promptsp;
 
 #define SHELLSET_OPTIONS L"abefhmnuvxC"
 
 extern const struct xoption
     *const shell_long_options, *const set_long_options, *const help_option;
+extern int setoptindex;
 
-extern void set_option(wchar_t c);
+extern void set_option(void);
+extern void set_single_option(wchar_t c);
 extern _Bool set_long_option(const wchar_t *s)
     __attribute__((nonnull));
 extern wchar_t *get_hyphen_parameter(void)
