@@ -224,15 +224,15 @@ void redirsfree(redir_T *r)
 
 /********** Auxiliary Functions for Parser **********/
 
-static inline bool is_name_char(wchar_t c)
+static bool is_portable_name_char(wchar_t c)
     __attribute__((const));
 static wchar_t *skip_name(const wchar_t *s)
     __attribute__((pure,nonnull));
 
 
-/* Checks if the specified character can be used in a variable name.
+/* Checks if the specified character can be used in a portable variable name.
  * Returns true for a digit. */
-bool is_name_char(wchar_t c)
+bool is_portable_name_char(wchar_t c)
 {
     switch (c) {
     case L'0':  case L'1':  case L'2':  case L'3':  case L'4':
@@ -247,12 +247,19 @@ bool is_name_char(wchar_t c)
     case L'M':  case L'N':  case L'O':  case L'P':  case L'Q':  case L'R':
     case L'S':  case L'T':  case L'U':  case L'V':  case L'W':  case L'X':
     case L'Y':  case L'Z':  case L'_':
-	return true;
+       return true;
     default:
-	return false;
+       return false;
     }
 }
 
+/* Checks if the specified character can be used in a variable name.
+ * Returns true for a digit. */
+bool is_name_char(wchar_t c)
+{
+    return c == L'_' || iswalnum(c);
+}
+
 /* Skips an identifier at the head of the specified string and returns a
  * pointer to the character right after the identifier in the string. */
 wchar_t *skip_name(const wchar_t *s)
@@ -263,9 +270,19 @@ wchar_t *skip_name(const wchar_t *s)
     return (wchar_t *) s;
 }
 
+/* Checks if the specified string constitutes a valid identifier. */
+bool is_name(const wchar_t *s)
+{
+    return s[0] != L'\0' && skip_name(s)[0] == L'\0';
+}
+
 /* Returns true iff the string is a shell reserved word. */
 bool is_keyword(const wchar_t *s)
 {
+    /* List of keywords:
+     *    case do done elif else esac fi for if in then until while { } !
+     * The following words are currently not keywords:
+     *    [[ ]] function select */
     switch (s[0]) {
 	case L'c':
 	    return s[1] == L'a' && s[2] == L's' && s[3] == L'e' && s[4]== L'\0';
@@ -322,6 +339,7 @@ static void serror(const char *restrict format, ...)
 static inline int read_more_input(void);
 static inline void line_continuation(size_t index);
 static void ensure_buffer(size_t n);
+static size_t count_name_length(bool isnamechar(wchar_t c));
 static void skip_blanks_and_comment(void);
 static bool skip_to_next_token(void);
 static void next_line(void);
@@ -541,7 +559,7 @@ void serror(const char *restrict format, ...)
 {
     va_list ap;
 
-    if (cinfo->print_errmsg) {
+    if (cinfo->print_errmsg && cinfo->lastinputresult != 1) {
 	if (cinfo->filename)
 	    fprintf(stderr, "%s:%lu: ", cinfo->filename, cinfo->lineno);
 	fprintf(stderr, gt("syntax error: "));
@@ -599,7 +617,7 @@ void ensure_buffer(size_t n)
 	    line_continuation(index);
 	    if (cinfo->lastinputresult != 0)
 		return;
-	    /* falls thru! */
+	    break;
 	default:
 	    index++;
 	    break;
@@ -607,6 +625,20 @@ void ensure_buffer(size_t n)
     }
 }
 
+/* Returns the length of the name under the current index.
+ * Whether a character can be part of the name is determined by `isnamechar'.
+ * Processes line continuations and reads enough lines so that the variable/alias
+ * name under the current index is fully available. */
+size_t count_name_length(bool isnamechar(wchar_t c))
+{
+    size_t savecindex = cindex, newcindex;
+    while (ensure_buffer(1), isnamechar(cbuf.contents[cindex]))
+	cindex++;
+    newcindex = cindex;
+    cindex = savecindex;
+    return newcindex - savecindex;
+}
+
 /* Advances `cindex' to skip blank characters, comments and line continuations.
  * This function may call `read_more_input' if needed. `cindex' is advanced to
  * the next non-blank character. If a comment is skipped, `cindex' will point to
@@ -665,7 +697,7 @@ void next_line(void)
 
     for (size_t i = 0; i < pending_heredocs.length; i++)
 	read_heredoc_contents(pending_heredocs.contents[i]);
-    pl_clear(&pending_heredocs);
+    pl_clear(&pending_heredocs, 0);
 }
 
 /* Checks if the specified character is a token separator */
@@ -952,9 +984,6 @@ command_T *parse_command(void)
     result->c_type = CT_SIMPLE;
     redirlastp = parse_assignments_and_redirects(result);
     result->c_words = parse_words_and_redirects(redirlastp, true);
-    /* `tryparse_function' removes all line continuations in the command word,
-     * so we don't have to worry about line continuations in alias substitution.
-     * But line continuations aren't handled properly for other words. */
 
     ensure_buffer(1);
     if (cbuf.contents[cindex] == L'(')
@@ -978,7 +1007,8 @@ redir_T **parse_assignments_and_redirects(command_T *c)
     while (ensure_buffer(1),
 	    !is_command_delimiter_char(cbuf.contents[cindex])) {
 #if YASH_ENABLE_ALIAS
-	substitute_alias(&cbuf, cindex, caliases, false);
+	size_t len = count_name_length(is_alias_name_char);
+	substitute_alias(&cbuf, cindex, len, caliases, false);
 	skip_blanks_and_comment();
 #endif
 	if ((redir = tryparse_redirect())) {
@@ -1011,7 +1041,8 @@ void **parse_words_and_redirects(redir_T **redirlastp, bool first)
 	    !is_command_delimiter_char(cbuf.contents[cindex])) {
 #if YASH_ENABLE_ALIAS
 	if (!first) {
-	    substitute_alias(&cbuf, cindex, caliases, true);
+	    size_t len = count_name_length(is_alias_name_char);
+	    substitute_alias(&cbuf, cindex, len, caliases, true);
 	    skip_blanks_and_comment();
 	}
 #endif
@@ -1037,10 +1068,8 @@ void parse_redirect_list(redir_T **lastp)
     for (;;) {
 #if YASH_ENABLE_ALIAS
 	if (!posixly_correct) {
-	    wchar_t *nameend;
-	    while (*(nameend = skip_name(cbuf.contents + cindex)) == L'\0'
-		    && read_more_input() == 0);
-	    substitute_alias(&cbuf, cindex, caliases, true);
+	    size_t len = count_name_length(is_alias_name_char);
+	    substitute_alias(&cbuf, cindex, len, caliases, true);
 	}
 #endif
 
@@ -1056,18 +1085,16 @@ void parse_redirect_list(redir_T **lastp)
  * Otherwise, does nothing and returns NULL. */
 assign_T *tryparse_assignment(void)
 {
-    size_t namelen;
+    if (iswdigit(cbuf.contents[cindex]))
+	return NULL;
 
-    do
-	namelen = skip_name(cbuf.contents + cindex) - (cbuf.contents + cindex);
-    while (cbuf.contents[cindex + namelen] == L'\0' && read_more_input() == 0);
+    size_t namelen = count_name_length(is_name_char);
     if (namelen == 0 || cbuf.contents[cindex + namelen] != L'=')
 	return NULL;
 
     assign_T *result = xmalloc(sizeof *result);
     result->next = NULL;
-    result->a_name = malloc_wcsntombs(cbuf.contents + cindex, namelen);
-    assert(result->a_name != NULL);
+    result->a_name = xwcsndup(cbuf.contents + cindex, namelen);
     cindex += namelen + 1;
 
     ensure_buffer(1);
@@ -1237,11 +1264,9 @@ wordunit_T *parse_word_to(aliastype_T type, bool testfunc(wchar_t c))
 	case noalias:
 	    break;
 	case globalonly:
-	    substitute_alias(&cbuf, cindex, caliases, true);
-	    skip_blanks_and_comment();
-	    break;
-	case anyalias:
-	    substitute_alias(&cbuf, cindex, caliases, false);
+	case anyalias:;
+	    size_t len = count_name_length(is_alias_name_char);
+	    substitute_alias(&cbuf, cindex, len, caliases, type == globalonly);
 	    skip_blanks_and_comment();
 	    break;
     }
@@ -1396,26 +1421,17 @@ wordunit_T *tryparse_paramexp_raw(void)
 	    namelen = 1;
 	    goto success;
     }
-    if (!is_name_char(cbuf.contents[cindex]))
+    if (!is_portable_name_char(cbuf.contents[cindex]))
 	goto fail;
-
-findnamelen:
-    namelen = skip_name(cbuf.contents + cindex) - (cbuf.contents + cindex);
-    if (cbuf.contents[cindex + namelen] == L'\\' &&
-	    cbuf.contents[cindex + namelen + 1] == L'\n') {
-	line_continuation(cindex + namelen);
-	goto findnamelen;
-    }
-    if (namelen == 0) {
-	assert(L'0' <= cbuf.contents[cindex] && cbuf.contents[cindex] <= L'9');
-	namelen++;
-    }
+    if (iswdigit(cbuf.contents[cindex]))
+	namelen = 1;
+    else
+	namelen = count_name_length(is_portable_name_char);
 
 success:
     pe = xmalloc(sizeof *pe);
     pe->pe_type = PT_NONE;
-    pe->pe_name = malloc_wcsntombs(cbuf.contents + cindex, namelen);
-    assert(pe->pe_name != NULL);
+    pe->pe_name = xwcsndup(cbuf.contents + cindex, namelen);
     pe->pe_start = pe->pe_end = pe->pe_match = pe->pe_subst = NULL;
 
     wordunit_T *result = xmalloc(sizeof *result);
@@ -1444,13 +1460,20 @@ wordunit_T *parse_paramexp_in_brase(void)
 
     /* parse PT_NUMBER */
     ensure_buffer(3);
-    if (cbuf.contents[cindex] == L'#'
-	    && cbuf.contents[cindex + 1] != L'}'
-	    && (cbuf.contents[cindex + 1] != L'#'
-		|| (cbuf.contents[cindex + 1] != L'\0'
-		    && cbuf.contents[cindex + 2] == L'}'))) {
-	pe->pe_type |= PT_NUMBER;
-	cindex++;
+    if (cbuf.contents[cindex] == L'#') {
+	switch (cbuf.contents[cindex + 1]) {
+	    case L'\0': case L' ':
+	    case L'+':  case L'=':  case L':':  case L'/':  case L'%':
+		break;
+	    case L'-':  case L'?':  case L'#':
+		if (cbuf.contents[cindex + 2] != L'}')
+		    break;
+		/* falls thru! */
+	    default:
+		pe->pe_type |= PT_NUMBER;
+		cindex++;
+		break;
+	}
     }
 
     /* parse nested expansion */
@@ -1458,8 +1481,11 @@ wordunit_T *parse_paramexp_in_brase(void)
     if (!posixly_correct && cbuf.contents[cindex] == L'{') {
 	pe->pe_type |= PT_NEST;
 	pe->pe_nest = parse_paramexp_in_brase();
-    } else if (!posixly_correct &&
-	    (cbuf.contents[cindex] == L'$' || cbuf.contents[cindex] == L'`')) {
+    } else if (!posixly_correct
+	    && (cbuf.contents[cindex] == L'`'
+		|| (cbuf.contents[cindex] == L'$'
+		    && (cbuf.contents[cindex + 1] == L'{'
+			|| cbuf.contents[cindex + 1] == L'(')))) {
 	size_t neststartindex = cindex;
 	pe->pe_nest = parse_special_word_unit();
 	if (cindex != neststartindex)
@@ -1483,7 +1509,7 @@ parse_name:;
 	    goto fail;
 	}
 make_name:
-	pe->pe_name = malloc_wcsntombs(
+	pe->pe_name = xwcsndup(
 		cbuf.contents + namestartindex, cindex - namestartindex);
 	assert(pe->pe_name != NULL);
     }
@@ -1708,6 +1734,8 @@ wordunit_T *tryparse_arith(void)
     for (;;) {
 	ensure_buffer(1);
 	switch (cbuf.contents[cindex]) {
+	case L'\0':
+	    goto fail;
 	case L'\\':
 	    if (cbuf.contents[cindex + 1] == L'\n') {
 		line_continuation(cindex);
@@ -1925,9 +1953,7 @@ command_T *parse_for(void)
 	serror(Ngt("no identifier after `for'"));
     else if (*skip_name(name) != L'\0')
 	serror(Ngt("`%ls' is not valid identifier"), name);
-    result->c_forname = realloc_wcstombs(name);
-    if (!result->c_forname)
-	serror(Ngt("cannot convert wide characters into multibyte characters"));
+    result->c_forname = name;
 
     skip_to_next_token();
     ensure_buffer(3);
@@ -2100,23 +2126,13 @@ command_T *tryparse_function(void)
     size_t savecindex = cindex;
     unsigned long lineno = cinfo->lineno;
 
-    wchar_t *nameend;
-    size_t namelen;
-
-    /* parse function name */
-readname:
-    while (*(nameend = skip_name(cbuf.contents + cindex)) == L'\0'
-	    && read_more_input() == 0);
-    namelen = nameend - (cbuf.contents + cindex);
-    if (namelen == 0 || !is_token_delimiter_char(*nameend)) {
-	ensure_buffer(2);
-	if (nameend[0] == L'\\' && nameend[1] == L'\n') {
-	    line_continuation(nameend - cbuf.contents);
-	    goto readname;
-	}
+    if (iswdigit(cbuf.contents[cindex]))
 	goto fail;
-    }
+
+    size_t namelen = count_name_length(is_name_char);
     cindex += namelen;
+    if (namelen == 0 || !is_token_delimiter_char(cbuf.contents[cindex]))
+	goto fail;
     skip_blanks_and_comment();
 
     /* parse parentheses */
@@ -2144,14 +2160,8 @@ readname:
     result->c_type = CT_FUNCDEF;
     result->c_lineno = lineno;
     result->c_redirs = NULL;
-    result->c_funcname = realloc_wcstombs(
-	    xwcsndup(cbuf.contents + savecindex, namelen));
+    result->c_funcname = xwcsndup(cbuf.contents + savecindex, namelen);
     result->c_funcbody = body;
-    if (!result->c_funcname) {
-	cerror = true;
-	xerror(0, Ngt("cannot convert wide characters "
-		    "into multibyte characters"));
-    }
     return result;
 
 fail:
@@ -2535,7 +2545,7 @@ void print_command_content(xwcsbuf_T *restrict buf, const command_T *restrict c)
 	break;
     case CT_FOR:
 	wb_cat(buf, L"for ");
-	wb_mbscat(buf, c->c_forname);
+	wb_cat(buf, c->c_forname);
 	if (c->c_forwords) {
 	    wb_cat(buf, L" in");
 	    for (void **w = c->c_forwords; *w; w++) {
@@ -2562,7 +2572,7 @@ void print_command_content(xwcsbuf_T *restrict buf, const command_T *restrict c)
 	wb_cat(buf, L"esac ");
 	break;
     case CT_FUNCDEF:
-	wb_mbscat(buf, c->c_funcname);
+	wb_cat(buf, c->c_funcname);
 	wb_cat(buf, L" () ");
 	print_commands(buf, c->c_funcbody);
 	break;
@@ -2592,7 +2602,7 @@ void print_caseitems(xwcsbuf_T *restrict buf, const caseitem_T *restrict i)
 void print_assigns(xwcsbuf_T *restrict buf, const assign_T *restrict a)
 {
     while (a) {
-	wb_mbscat(buf, a->a_name);
+	wb_cat(buf, a->a_name);
 	wb_wccat(buf, L'=');
 	switch (a->a_type) {
 	    case A_SCALAR:
@@ -2689,7 +2699,7 @@ void print_paramexp(xwcsbuf_T *restrict buf, const paramexp_T *restrict p)
     if (p->pe_type & PT_NEST)
 	print_word(buf, p->pe_nest);
     else
-	wb_mbscat(buf, p->pe_name);
+	wb_cat(buf, p->pe_name);
     if (p->pe_start) {
 	wb_wccat(buf, L'[');
 	print_word(buf, p->pe_start);
