@@ -46,7 +46,7 @@ extern _Bool is_same_file(const char *path1, const char *path2)
 
 extern wchar_t *canonicalize_path(const wchar_t *path)
     __attribute__((nonnull,malloc,warn_unused_result));
-extern _Bool is_canonicalized(const wchar_t *path)
+extern _Bool is_normalized_path(const wchar_t *path)
     __attribute__((nonnull));
 
 extern char *xgetcwd(void)
