@@ -22,6 +22,7 @@
 #include <stddef.h>
 
 
+extern bool process_exists(pid_t pid);
 extern const char *get_signal_name(int signum)
     __attribute__((const));
 extern int get_signal_number(const char *name)
@@ -51,6 +52,7 @@ extern _Bool have_unhandled_traps(void);
 extern int handle_traps(void);
 extern void execute_exit_trap(void);
 extern void clear_traps(void);
+extern void reset_sigwinch(void);
 
 extern int trap_builtin(int argc, void **argv)
     __attribute__((nonnull));
