@@ -58,6 +58,8 @@ static void execute_profile(void);
 static void execute_rcfile(const wchar_t *rcfile);
 static void print_help(void);
 static void print_version(void);
+static bool input_is_interactive_terminal(const parseinfo_T *pinfo)
+    __attribute__((nonnull));
 
 /* the process ID of the shell */
 pid_t shell_pid;
@@ -126,7 +128,7 @@ int main(int argc, char **argv)
 	    (opt = xgetopt_long(wargv,
 		    L"+*cilo:sV" SHELLSET_OPTIONS,
 		    shell_long_options,
-		    NULL))) {
+		    &setoptindex))) {
 	switch (opt) {
 	case L'c':
 	    if (xoptopt != L'-') {
@@ -143,7 +145,11 @@ int main(int argc, char **argv)
 	case L'l':
 	    is_login_shell = (xoptopt == L'-');
 	    break;
+	case L'L':
+	    set_option();
+	    break;
 	case L'o':
+	    do_job_control_set |= wcscmp(xoptarg, L"monitor") == 0;
 	    if (!set_long_option(xoptarg)) {
 		xerror(0, Ngt("%lco %ls: invalid option"),
 			(wint_t) xoptopt, xoptarg);
@@ -180,7 +186,7 @@ int main(int argc, char **argv)
 	    do_job_control_set = true;
 	    /* falls thru! */
 	default:
-	    set_option(opt);
+	    set_single_option(opt);
 	    break;
 	}
     }
@@ -200,6 +206,8 @@ int main(int argc, char **argv)
     if (wargv[xoptind] && wcscmp(wargv[xoptind], L"-") == 0)
 	xoptind++;
 
+    /* `shell_pid' must be initialized after the options have been parsed.
+     * This is required for the `set_monitor_option' function to work. */
     shell_pid = getpid();
     initial_pgid = shell_pgid = getpgrp();
     init_cmdhash();
@@ -251,6 +259,14 @@ int main(int argc, char **argv)
 	}
     }
 
+#if YASH_ENABLE_LINEEDIT
+    /* enable lineedit if interactive and connected to a terminal */
+    if (is_interactive)
+	if (!is_interactive_set
+		|| (isatty(STDIN_FILENO) && isatty(STDERR_FILENO)))
+	    shopt_lineedit = shopt_vi;
+#endif
+
     is_interactive_now = is_interactive;
     if (!do_job_control_set)
 	do_job_control = is_interactive;
@@ -265,9 +281,6 @@ int main(int argc, char **argv)
 	if (!norcfile)
 	    execute_rcfile(rcfile);
     }
-#if YASH_ENABLE_HISTORY
-    init_history();
-#endif
 
     if (shopt_read_arg) {
 	exec_wcs(command, posixly_correct ? "sh -c" : "yash -c", true);
@@ -309,7 +322,7 @@ static void execute_rcfile(const wchar_t *rcfile)
     if (!is_interactive)
 	return;
     if (posixly_correct) {
-	const wchar_t *env = getvar(VAR_ENV);
+	const wchar_t *env = getvar(L VAR_ENV);
 	if (!env)
 	    return;
 	wpath = parse_and_expand_string(env, "$ENV", false);
@@ -354,7 +367,7 @@ void exit_shell_with_status(int status)
 	execute_exit_trap();
 #if YASH_ENABLE_HISTORY
 	if (is_interactive_now)
-	    write_history(NULL, false);
+	    finalize_history();
 #endif
     }
     finalize_shell();
@@ -423,6 +436,9 @@ void print_version(void)
 #if YASH_ENABLE_HISTORY
 		" * history\n"
 #endif
+#if YASH_ENABLE_LINEEDIT
+		" * lineedit\n"
+#endif
 #if YASH_ENABLE_PRINTF
 		" * printf/echo\n"
 #endif
@@ -493,7 +509,6 @@ void exec_mbs(const char *code, const char *name, bool finally_exit)
 	.input = input_mbs,
 	.inputinfo = &iinfo,
 	.intrinput = false,
-	.inputisatty = false,
 	.lastinputresult = 0,
     };
     memset(&iinfo.state, 0, sizeof iinfo.state);  // initialize the shift state
@@ -517,7 +532,6 @@ void exec_wcs(const wchar_t *code, const char *name, bool finally_exit)
 	.input = input_wcs,
 	.inputinfo = &iinfo,
 	.intrinput = false,
-	.inputisatty = false,
 	.lastinputresult = 0,
     };
 
@@ -545,11 +559,9 @@ void exec_input(FILE *f, const char *name, bool intrinput, bool finally_exit)
 	rlinfo.type = 1;
 	pinfo.input = input_readline;
 	pinfo.inputinfo = &rlinfo;
-	pinfo.inputisatty = isatty(fileno(f));
     } else {
 	pinfo.input = (f == stdin) ? input_stdin : input_file;
 	pinfo.inputinfo = f;
-	pinfo.inputisatty = false;
     }
     parse_and_exec(&pinfo, finally_exit);
 }
@@ -571,10 +583,14 @@ void parse_and_exec(parseinfo_T *pinfo, bool finally_exit)
 	nextforceexit = false;
 	if (return_pending()) {
 	    reset_execinfo();
-	    if (pinfo->intrinput)
+	    if (pinfo->intrinput) {
 		xerror(0, Ngt("return: not in function or sourced file"));
-	    else
-		goto finish;
+	    } else if (finally_exit) {
+		wchar_t argv0[] = L"return";
+		exit_builtin(1, (void *[]) { argv0, NULL });
+		continue;
+	    } else
+		goto out;
 	}
 
 	switch (read_and_parse(pinfo, &commands)) {
@@ -590,13 +606,12 @@ void parse_and_exec(parseinfo_T *pinfo, bool finally_exit)
 		}
 		break;
 	    case EOF:
-		if (shopt_ignoreeof && pinfo->inputisatty) {
+		if (shopt_ignoreeof && input_is_interactive_terminal(pinfo)) {
 		    fprintf(stderr, gt("Use `exit' to leave the shell.\n"));
 		    break;
 		}
 		if (!executed)
 		    laststatus = Exit_SUCCESS;
-finish:
 		if (finally_exit) {
 		    wchar_t argv0[] = L"EOF";
 		    exit_builtin(1, (void *[]) { argv0, NULL });
@@ -617,9 +632,20 @@ finish:
      * We don't set it at first because the value of "$?" gets wrong in
      * execution. */
 out:
+    assert(!finally_exit);
     load_execinfo(ei);
 }
 
+bool input_is_interactive_terminal(const parseinfo_T *pinfo)
+{
+    struct input_readline_info *ir;
+
+    if (!pinfo->intrinput)
+	return false;
+    ir = pinfo->inputinfo;
+    return isatty(fileno(ir->fp));
+}
+
 
 /********** Builtins **********/
 
@@ -691,8 +717,8 @@ const char exit_help[] = Ngt(
 "If <n> is not specified, it defaults to the exit status of the last executed\n"
 "command. <n> should be between 0 and 255 inclusive.\n"
 "If the shell is interactive and you have any stopped jobs, the shell prints\n"
-"a warning message and does not exit. Use the -f option or use `exit' twice \n"
-"in a row to avoid the warning and really exit.\n"
+"a warning message and does not exit. Use the -f (--force) option or use\n"
+"`exit' twice in a row to avoid the warning and really exit.\n"
 );
 
 /* The "suspend" builtin */
